{
GPC-Test is public domain software written by Richard B. Winston
(rbwinst@usgs.gov).
It relies on GPC (http://www.cs.man.ac.uk/~toby/alan/software/ ,
http://www.cs.man.ac.uk/~toby/alan/software/gpc.html)
and Graphics32 (http://sourceforge.net/project/showfiles.php?group_id=112985&package_id=122211).

It also uses TQRbwZoomBox2 which is a public domain custom component
included with the source code of GPC-Test.

To compile GPC-Test first, you must first install Graphics32
and then TQRbwZoomBox2 in Delphi.

Neither Graphics32 nor GPC are in the public domain. Graphics32 is covered
by the MPL 1.1 license. GPC is free for non-commercial use. 
}

unit frmMainUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, ZoomBox2, GR32, GR32_Layers, GR32_Polygons, GPC,
  Mask, JvExMask, JvToolEdit;

type
  TfrmMain = class(TForm)
    ZoomBox: TQRbwZoomBox2;
    fePoly1: TJvFilenameEdit;
    fePoly2: TJvFilenameEdit;
    rgClipOperation: TRadioGroup;
    rgOutputType: TRadioGroup;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure fePolyChange(Sender: TObject);
    procedure rgClipOperationClick(Sender: TObject);
    procedure ZoomBoxResize(Sender: TObject);
    procedure rgOutputTypeClick(Sender: TObject);
  private
    Polygon1: Tgpc_polygon;
    Polygon2: Tgpc_polygon;
    ResultPolygon: Tgpc_polygon;
    Tristrip: Tgpc_tristrip;
    FPositionedLayer: TPositionedLayer;
    FPaintingLayer: Boolean;
    FBitMap32: TBitmap32;
    FDrawing: Boolean;
    FShouldUpdate: Boolean;
    // Display code
    procedure PaintLayer(Sender: TObject; Buffer: TBitmap32);
    procedure PaintBitmap(Sender: TObject);
    procedure DrawContours;
    procedure DrawAPolygon(APolygon: Tgpc_polygon; FillColor: TColor32);
    procedure DrawTristrip;

    procedure ReadPolygon(var Polygon: Tgpc_polygon; FileName: string);
    procedure GetMinMax(Polygon: Tgpc_polygon; var FirstFound: Boolean;
      var MaxY, MinY, MaxX, MinX: Double);
    procedure UpdateScale;
    procedure UpdateClipOperation;
    procedure ReadFiles;
    { Private declarations }
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

uses
  Math, BigCanvasMethods;

{$R *.dfm}

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  FBitMap32 := TBitmap32.Create;

  // ZoomBox.Image32.Layers is a TCollection.
  // You can have it hold different sorts of TCustomLayer
  // by passing it a descendant of TCustomLayer.
  // Here we use TPositionedLayer.
  FPositionedLayer := ZoomBox.Image32.Layers.Add(TPositionedLayer) as
    TPositionedLayer;

  // Assign an event handler for the OnPaint event.
  FPositionedLayer.OnPaint := PaintLayer;

  // fePolyChange will be called automatically when
  // fePoly1.FileName or fePoly2.FileName changes.
  fePoly1.FileName := IncludeTrailingPathDelimiter(ExtractFileDir(
    Application.ExeName)) + 'britain.gpf';
  fePoly2.FileName := IncludeTrailingPathDelimiter(ExtractFileDir(
    Application.ExeName)) + 'arrows.gpf';

  ZoomBox.Image32.Invalidate;
end;

procedure TfrmMain.fePolyChange(Sender: TObject);
begin
  // This is the event handler for both
  // the fePoly1.OnChange and fePoly2.OnChange events.
  ReadFiles;
end;

procedure TfrmMain.ReadFiles;
begin
  if FileExists(fePoly1.FileName) and FileExists(fePoly2.FileName) then
  begin
    ReadPolygon(Polygon1, fePoly1.FileName);
    ReadPolygon(Polygon2, fePoly2.FileName);
    UpdateClipOperation;
    UpdateScale;
  end;
end;


procedure TfrmMain.ReadPolygon(var Polygon: Tgpc_polygon; FileName: string);
const
  FFalse = 0;
var
  AFile: TextFile;
begin
  AssignFile(AFile, FileName);
  try
    Reset(AFile);
    gpc_free_polygon(@Polygon);
    gpc_read_polygon(AFile, FFalse, @Polygon);
  finally
    CloseFile(AFile);
  end;
end;

procedure TfrmMain.UpdateClipOperation;
begin
  if FileExists(fePoly1.FileName) and FileExists(fePoly2.FileName) then
  begin
    try
      Screen.Cursor := crHourGlass;
      case rgOutputType.ItemIndex of
        0:
          begin
            gpc_free_polygon(@ResultPolygon);
            gpc_polygon_clip(Tgpc_op(rgClipOperation.ItemIndex),
              @Polygon1, @Polygon2, @ResultPolygon);
          end;
        1:
          begin
            gpc_free_tristrip(@Tristrip);
            gpc_tristrip_clip(Tgpc_op(rgClipOperation.ItemIndex),
              @Polygon1, @Polygon2, @Tristrip);
          end;
        else Assert(False);
      end;
      FShouldUpdate := True;
      ZoomBox.Image32.Invalidate;
    finally
      Screen.Cursor := crDefault;
    end;
  end;
end;

procedure TfrmMain.UpdateScale;
var
  FirstFound: Boolean;
  MaxY: Double;
  MinY: Double;
  MaxX: Double;
  MinX: Double;
begin
  FirstFound := False;
  MaxY := 0;
  MinY := 0;
  MaxX := 0;
  MinX := 0;
  GetMinMax(Polygon1, FirstFound, MaxY, MinY, MaxX, MinX);
  GetMinMax(Polygon2, FirstFound, MaxY, MinY, MaxX, MinX);
  if FirstFound then
  begin
    ZoomBox.Magnification := Min((ZoomBox.Width - 1) / (MaxX - MinX),
      (ZoomBox.Height - 10) / (MaxY - MinY));
    ZoomBox.OriginX := MinX;
    ZoomBox.OriginY := MinY;
    ZoomBox.OriginX := ZoomBox.X(-5);
    ZoomBox.OriginY := ZoomBox.Y(ZoomBox.Height + 5);
    FShouldUpdate := True;
    ZoomBox.Image32.Invalidate;
  end;
end;

procedure TfrmMain.GetMinMax(Polygon: Tgpc_polygon; var FirstFound: Boolean;
  var MaxY, MinY, MaxX, MinX: Double);
var
  AVertex: Tgpc_vertex;
  VertexIndex: Integer;
  ContourIndex: Integer;
begin
  for ContourIndex := 0 to Polygon.num_contours - 1 do
  begin
    with Polygon.contour[ContourIndex] do
    begin
      for VertexIndex := 0 to num_vertices - 1 do
      begin
        AVertex := vertex[VertexIndex];
        if FirstFound then
        begin
          if MinX > AVertex.x then
          begin
            MinX := AVertex.x;
          end;
          if MaxX < AVertex.x then
          begin
            MaxX := AVertex.x;
          end;
          if MinY > AVertex.y then
          begin
            MinY := AVertex.y;
          end;
          if MaxY < AVertex.y then
          begin
            MaxY := AVertex.y;
          end;
        end
        else
        begin
          FirstFound := True;
          MinX := AVertex.x;
          MaxX := AVertex.x;
          MinY := AVertex.y;
          MaxY := AVertex.y;
        end;
      end;
    end;
  end;
end;

procedure TfrmMain.PaintLayer(Sender: TObject; Buffer: TBitmap32);
begin
  if FPaintingLayer then
  begin
    Exit;
  end;
  FPaintingLayer := True;
  try
    Buffer.BeginUpdate;
    try
      PaintBitmap(Sender);
      Buffer.Draw(0, 0, FBitMap32);
    finally
      Buffer.EndUpdate;
    end;
  finally
    FPaintingLayer := False;
  end;
end;

procedure TfrmMain.PaintBitmap(Sender: TObject);
begin
  if FDrawing then
    Exit;
  FDrawing := True;
  try
    if (FBitMap32 <> nil) then
    begin
      // if you need to redraw the FBitMap32,
      // change the cursor to an hourglass,
      // redraw, and change the cursor back.
      if FShouldUpdate then
      begin
        FShouldUpdate := False;
        try
          DrawContours;
        except on EInvalidOp do
          begin
            ZoomBox.ZoomBy(0.01);
            FShouldUpdate := True;
            ZoomBox.Image32.Invalidate;
            Exit;
          end;
        end;
      end;
    end;
  finally
    FDrawing := False;
    if FShouldUpdate then
    begin
      ZoomBox.Image32.Invalidate;
    end;
  end;
end;


procedure TfrmMain.DrawContours;
begin
  FBitMap32.Free;
  FBitMap32 := nil;
  FBitMap32 := TBitmap32.Create;

  FBitMap32.Height := ZoomBox.Image32.Height;
  FBitMap32.Width := ZoomBox.Image32.Width;

  // draw a box around the drawing area
  DrawBigRectangle32(FBitMap32, clBlack32, clWhite32, 1.0, 0, 0,
    ZoomBox.ClientWidth - 1, ZoomBox.ClientHeight - 1);

  // draw the result of the clipping operation.
  case rgOutputType.ItemIndex of
    0:
      begin
        DrawAPolygon(ResultPolygon, clRed32);
      end;
    1:
      begin
        DrawTristrip;
      end;
    else Assert(False);
  end;
  // Draw the polygons.
  DrawAPolygon(Polygon1, clTransparent32);
  DrawAPolygon(Polygon2, clTransparent32);

end;

procedure TfrmMain.DrawAPolygon(APolygon: Tgpc_polygon; FillColor: TColor32);
var
  DummyPolygon: TPolygon32;
  ContourIndex: Integer;
  Points: TPointArray;
  VertexIndex: Integer;
  AVertex: Tgpc_vertex;
  ColorUsed: TColor32;
begin
  DummyPolygon := nil;
  for ContourIndex := 0 to APolygon.num_contours - 1 do
  begin
    with APolygon.contour[ContourIndex] do
    begin
      SetLength(Points, num_vertices + 1);
      for VertexIndex := 0 to num_vertices - 1 do
      begin
        AVertex := vertex[VertexIndex];
        Points[VertexIndex].X := ZoomBox.XCoord(AVertex.x);
        Points[VertexIndex].Y := ZoomBox.YCoord(AVertex.y);
      end;
      Points[num_vertices] := Points[0];

      if (APolygon.hole <> nil) and (APolygon.hole[ContourIndex] = 1) then
      begin
        ColorUsed := clWhite32;
      end
      else
      begin
        ColorUsed := FillColor;
      end;

      DrawBigPolygon32(FBitmap32, clBlack32, ColorUsed, 1, Points,
        DummyPolygon, False, True);
    end;
  end;
end;

procedure TfrmMain.DrawTristrip;
var
  DummyPolygon: TPolygon32;
  V3: Tgpc_vertex;
  V2: Tgpc_vertex;
  V1: Tgpc_vertex;
  VertexIndex: Integer;
  Strip: Tgpc_vertex_list;
  StripIndex: Integer;
  Points: TPointArray;
begin
  DummyPolygon := nil;
  SetLength(Points, 4);
  for StripIndex := 0 to Tristrip.num_strips - 1 do
  begin
    Strip := Tristrip.Strip[StripIndex];
    for VertexIndex := 0 to Strip.num_vertices - 3 do
    begin
      V1 := Strip.vertex[VertexIndex];
      V2 := Strip.vertex[VertexIndex + 1];
      V3 := Strip.vertex[VertexIndex + 2];
      Points[0].X := ZoomBox.XCoord(V1.x);
      Points[0].Y := ZoomBox.YCoord(V1.y);
      Points[1].X := ZoomBox.XCoord(V2.x);
      Points[1].Y := ZoomBox.YCoord(V2.y);
      Points[2].X := ZoomBox.XCoord(V3.x);
      Points[2].Y := ZoomBox.YCoord(V3.y);
      Points[3] := Points[0];
      DrawBigPolygon32(FBitmap32, clLime32, clLime32, 1,
        Points, DummyPolygon, False, True);
    end;
  end;
end;

procedure TfrmMain.rgClipOperationClick(Sender: TObject);
begin
  UpdateClipOperation;
end;

procedure TfrmMain.rgOutputTypeClick(Sender: TObject);
begin
  UpdateClipOperation;
end;

procedure TfrmMain.ZoomBoxResize(Sender: TObject);
begin
  UpdateScale;
end;

procedure TfrmMain.FormDestroy(Sender: TObject);
begin
  FPositionedLayer.Free;
  FBitMap32.Free;
  gpc_free_polygon(@Polygon1);
  gpc_free_polygon(@Polygon2);
  gpc_free_polygon(@ResultPolygon);
  gpc_free_tristrip(@Tristrip);
end;

end.
